# coding: utf-8
from ..util import UnicodeMixin

from .schemas import ProjectSchema, EmptySchema, ProjectChrootSchema, BuildSchema, BuildTaskSchema, MockChrootSchema, \
    ProjectCreateSchema


class Link(UnicodeMixin):
    def __init__(self, role, href):
        self.role = role
        self.href = href

    def __unicode__(self):
        return u"<Link: role: {0}, href: {1}".format(self.role, self.href)

    @classmethod
    def from_dict(cls, data_dict):
        retval = {}
        for role_name, definition in data_dict.items():
            retval[role_name] = cls(role_name, definition["href"])
        return retval


# pylint: disable=E1101
class Entity(UnicodeMixin):
    _schema = EmptySchema()

    def __init__(self, **kwargs):
        for field in self._schema.fields.keys():
            setattr(self, field, kwargs.get(field))

    def to_dict(self):
        return self._schema.dump(self).data

    def to_json(self):
        return self._schema.dumps(self).data

    @classmethod
    def from_dict(cls, raw_dict):
        parsed = cls._schema.load(raw_dict)
        return cls(**parsed.data)


class ProjectEntity(Entity):
    _schema = ProjectSchema(strict=True)

    def __unicode__(self):
        return "<Project #{0}: {1}/{2}>".format(self.id, self.owner, self.name)


class ProjectCreateEntity(Entity):
    _schema = ProjectCreateSchema(strict=True)

    def __unicode__(self):
        return "<New project {0}/{1}>".format(self.owner, self.name)

class ProjectChrootEntity(Entity):
    _schema = ProjectChrootSchema(strict=True)

    def __unicode__(self):
        return "<Project chroot: {0}, additional " \
               "packages: {1}, comps size if any: {2}>"\
            .format(self.name, self.buildroot_pkgs, self.comps_len,)


class BuildEntity(Entity):
    _schema = BuildSchema(strict=True)

    def __unicode__(self):
        return "<Build #{0} state: {1}>".format(self.id, self.state)

    def is_finished(self):
        """
        Check is the build was finished

        :rtype: bool
        """
        return self.state in [
                "failed",
                "skipped",
                "succeeded"
        ]


class BuildTaskEntity(Entity):
    _schema = BuildTaskSchema(strict=True)

    def __unicode__(self):
        return "<Build task #{0}-{1}, state: {2}>".format(
            self.build_id, self.chroot_name, self.state
        )


class MockChrootEntity(Entity):
    _schema = MockChrootSchema(strict=True)

    def __unicode__(self):
        return "<Mock chroot: {0} is active: {1}>".format(
            self.name, self.is_active
        )



# coding: utf-8
from abc import abstractmethod, ABCMeta
import json
import os

from copr.client_v2.net_client import RequestError, MultiPartTuple
from copr.client_v2.schemas import ProjectCreateSchema
from .entities import ProjectChrootEntity, ProjectEntity, ProjectCreateEntity
from .resources import Project, OperationResult, ProjectList, ProjectChroot, ProjectChrootList, Build, BuildList, \
    MockChroot, MockChrootList, BuildTask, BuildTaskList


class AbstractHandle(object):
    """
    :param client: Should be used only to access other handlers
    :type client: copr.client_v2.client.HandlersProvider
    :type nc: copr.client_v2.net_client.NetClient
    """
    __metaclass__ = ABCMeta

    def __init__(self, client, nc, root_url):
        self.client = client
        self.nc = nc
        self.root_url = root_url

    @abstractmethod
    def get_base_url(self, *args, **kwargs):
        pass


class BuildHandle(AbstractHandle):
    def __init__(self, client, nc, root_url, builds_href):
        super(BuildHandle, self).__init__(client, nc, root_url)
        self.builds_href = builds_href
        self._base_url = "{0}{1}".format(self.root_url, builds_href)

    def get_base_url(self):
        return self._base_url

    def get_one(self, build_id):
        """ Retrieves builds object

        :param int build_id: id of the target build
        :rtype: :py:class:`~.resources.Build`
        """

        options = {"build_id": build_id}
        url = "{0}/{1}".format(self.get_base_url(), build_id)
        response = self.nc.request(url)
        return Build.from_response(
            handle=self,
            response=response,
            data_dict=response.json,
            options=options,
        )

    def get_list(self, project_id=None, owner=None, limit=None, offset=None):
        """ Retrieves builds object according to the given parameters

        :param owner: name of the project owner
        :param project_id: id of the project
        :param limit: limit number of builds
        :param offset: number of builds to skip

        :rtype: :py:class:`~.resources.BuildList`
        """
        options = {
            "project_id": project_id,
            "owner": owner,
            "limit": limit,
            "offset": offset
        }

        response = self.nc.request(self.get_base_url(), query_params=options)
        return BuildList.from_response(self, response, options)

    def cancel(self, build_entity):
        """ Cancels the given build

        :param build_entity: build entity to delete
        :type build_entity: :py:class:`~.copr.client_v2.entities.BuildEntity`

        :rtype: :py:class:`.OperationResult`
        """
        build_id = build_entity.id
        build_entity.state = "canceled"

        url = "{0}/{1}".format(self.get_base_url(), build_id)
        response = self.nc.request(url, data=build_entity.to_json(), method="PUT", do_auth=True)
        return OperationResult(self, response)

    def delete(self, build_id):
        """ Deletes the given build

        :param int build_id: build id to delete

        :rtype: :py:class:`.OperationResult`
        """

        url = "{0}/{1}".format(self.get_base_url(), build_id)
        response = self.nc.request(url, method="delete", do_auth=True)
        return OperationResult(self, response, expected_status=204)

    def _process_create_response(self, request_data, response):
        op_result = OperationResult(self, response, expected_status=201)
        if op_result.is_successful():
            build_response = self.nc.get(op_result.new_location)
            return Build.from_response(
                handle=self, response=build_response,
                data_dict=build_response.json
            )
        else:
            raise RequestError(
                "Got unexpected status code at create build request",
                url=self.get_base_url(),
                request_body=request_data, response=response
            )

    def create_from_url(self, project_id, srpm_url,
                        chroots=None, enable_net=True):

        """
        Creates new build using public url to the srpm file

        :param int project_id: id of the project where we want to submit new build
        :param str srpm_url: url to the source rpm
        :param list chroots: which chroots should be used during the build
        :param bool enable_net: allows to disable network access during the build, default: True

        :return: created build
        :rtype: :py:class:`~.resources.Build`
        """

        chroots = map(str, chroots or list())
        content = {
            "project_id": int(project_id),
            "srpm_url": str(srpm_url),
            "chroots": chroots,
            "enable_net": bool(enable_net)
        }
        data = json.dumps(content)
        response = self.nc.request(
            self.get_base_url(),
            data=data, method="POST", do_auth=True,
        )

        return self._process_create_response(data, response)

    def create_from_file(self, project_id, file_path=None,
                         file_obj=None, file_name=None,
                         chroots=None, enable_net=True):
        """
        Creates new build using srpm upload, please specify
        either ``file_path`` or (``file_obj``, ``file_name``    ).

        :param int project_id: id of the project where we want to submit new build

        :param str file_path: path to the srpm file

        :param file file_obj: file-like object to read from
        :param str file_name: name for the uploaded file

        :param list chroots: which chroots should be used during the build
        :param bool enable_net: allows to disable network access during the build, default: True

        :return: created build
        :rtype: :py:class:`~.resources.Build`
        """

        chroots = map(str, chroots or list())
        content = {
            "project_id": int(project_id),
            "chroots": chroots,
            "enable_net": bool(enable_net)
        }

        metadata = MultiPartTuple(
            "metadata", name=None,
            obj=json.dumps(content), content_type="application/json")

        if file_path is not None:
            with open(file_path, "rb") as f_obj:
                f_name = os.path.basename(f_obj.name)
                parts, response = self._do_upload_request(f_name, f_obj, metadata)
        elif file_obj is not None and file_name is not None:
            parts, response = self._do_upload_request(file_name, file_obj, metadata)
        else:
            raise RuntimeError("Please provide file_path or file_obj and file_name")

        return self._process_create_response(parts, response)

    def _do_upload_request(self, f_name, f_obj, metadata):
        srpm = MultiPartTuple("srpm", name=f_name, obj=f_obj,
                              content_type="appliction/x-rpm")
        parts = [metadata, srpm]
        response = self.nc.request_multipart(
            url=self.get_base_url(),
            method="POST",
            data_parts=parts,
            do_auth=True,
        )
        return parts, response

    def get_build_tasks_handle(self):
        """
        :rtype: BuildTasksHandle
        """
        return self.client.build_tasks

class BuildTaskHandle(AbstractHandle):

    def __init__(self, client, nc, root_url, build_tasks_href):
        super(BuildTaskHandle, self).__init__(client, nc, root_url)
        self.build_tasks_href = build_tasks_href
        self._base_url = "{0}{1}".format(self.root_url, build_tasks_href)

    def get_base_url(self):
        return self._base_url

    def get_list(self, owner=None, project_id=None, build_id=None,
                 state=None, offset=None, limit=None):

        """ Retrieves build tasks list according to the given parameters

        :param str owner: build tasks from the project owner by this user
        :param int project_id: get tasks only from this project,
            when used query parameter ``owner`` is ignored
        :param int build_id: get tasks only from this build,
            when used query parameters ``owner`` and ``project_id`` are ignored
        :param str state: get build tasks only with this state, allowed values:
            ``failed``, ``succeeded``, ``canceled``, ``running``,
            ``pending``, ``starting``, ``importing``
        :param int limit: limit number of projects
        :param int offset: number of projects to skip

        :rtype: :py:class:`~.resources.BuildTaskList`
        """
        options = {
            "owner": owner,
            "project_id": project_id,
            "build_id": build_id,
            "state": state,
            "limit": limit,
            "offset": offset
        }

        response = self.nc.request(self.get_base_url(), query_params=options)
        return BuildTaskList.from_response(self, response, options)

    def get_one(self, build_id, chroot_name):
        """ Retrieves single build task object


        :param int build_id: id of the build
        :param str chroot_name: name of the build chroot
        :rtype:  :py:class:`~.resources.BuildTask`
        """

        url = "{0}/{1}/{2}".format(self.get_base_url(), build_id, chroot_name)
        response = self.nc.request(url)
        return BuildTask.from_response(
            handle=self,
            response=response,
            data_dict=response.json,
        )


class ProjectHandle(AbstractHandle):

    def __init__(self, client, nc, root_url, projects_href):
        super(ProjectHandle, self).__init__(client, nc, root_url)
        self.projects_href = projects_href
        self._base_url = "{0}{1}".format(self.root_url, projects_href)

    def get_base_url(self):
        return self._base_url

    def get_list(self, search_query=None, owner=None, name=None, limit=None, offset=None):
        """ Retrieves projects object according to the given parameters

        :param str search_query: search projects with such string
        :param str owner: owner username
        :param str name: project name
        :param int limit: limit number of projects
        :param int offset: number of projects to skip

        :rtype: :py:class:`~.resources.ProjectList`
        """
        options = {
            "search_query": search_query,
            "owner": owner,
            "name": name,
            "limit": limit,
            "offset": offset
        }

        response = self.nc.request(self.get_base_url(), query_params=options)
        return ProjectList.from_response(self, response, options)

    def get_one(self, project_id):
        # todo: implement: , show_builds=False, show_chroots=False):
        """ Retrieves project object.

        :param int project_id: project identifier
        :rtype: :py:class:`~.resources.Project`
        """
        query_params = {
            # "show_builds": show_builds,
            # "show_chroots": show_chroots
        }

        url = "{0}/{1}".format(self.get_base_url(), project_id)
        response = self.nc.request(url, query_params=query_params)
        return Project.from_response(
            handle=self,
            response=response,
            data_dict=response.json,
            options=query_params
        )

    def create(
            self, name, owner, chroots, description=None, instructions=None,
            homepage=None, contact=None, disable_createrepo=None, build_enable_net=None,
            repos=None,
    ):
        """ Creates new project

        :param name: project name
        :param owner: username
        :param chroots: list of mock chroot to be used in project
        :param description:
        :param instructions:
        :param homepage:
        :param contact:
        :param bool disable_createrepo:
        :param bool build_enable_net:
        :param repos: list of additional repos enabled for builds

        :rtype: :py:class:`~.resources.Project`
        """

        new_entity = ProjectCreateEntity(
            owner=owner,
            name=name,
            chroots=chroots,
            description=description,
            instructions=instructions,
            homepage=homepage,
            contact=contact,
            disable_createrepo=disable_createrepo,
            build_enable_net=build_enable_net,
            repos=repos
        )

        url = self.get_base_url()
        request_data = new_entity.to_json()
        response = self.nc.request(url, method="post", data=request_data, do_auth=True)
        op_result = OperationResult(self, response, expected_status=201)
        if op_result.is_successful():
            response = self.nc.get(op_result.new_location)
            return Project.from_response(
                handle=self, response=response, data_dict=response.json)
        else:
            raise RequestError(
                "Got unexpected status code at create build request",
                url=self.get_base_url(),
                request_body=request_data, response=response
            )

    def update(self, project_entity):
        """ Updates project.

        :param project_entity: project entity to use for update
        :type project_entity: :py:class:`~.ProjectEntity`
        :rtype: OperationResult
        """
        url = "{0}/{1}".format(self.get_base_url(), project_entity.id)
        data = project_entity.to_json()

        response = self.nc.request(url, method="put", data=data, do_auth=True)
        return OperationResult(self, response)

    def delete(self, project_id):
        """ Deletes project.

        :param int project_id: project identifier
        :rtype: OperationResult
        """
        url = "{0}/{1}".format(self.get_base_url(), project_id)
        response = self.nc.request(url, method="delete", do_auth=True)
        return OperationResult(self, response, expected_status=204)

    def get_builds_handle(self):
        """
        :rtype: BuildHandle
        """
        return self.client.builds

    def get_build_tasks_handle(self):
        """
        :rtype: BuildTasksHandle
        """
        return self.client.build_tasks

    def get_project_chroots_handle(self):
        """
        :rtype: ProjectChrootHandle
        """
        return self.client.project_chroots


class ProjectChrootHandle(AbstractHandle):

    def get_base_url(self, project, **kwargs):
        """
        :type project: copr.client_v2.resources.Project
        """
        return "{0}{1}".format(self.root_url, project.get_href_by_name("chroots"))

    def get_one(self, project, name):
        """ Retrieves project chroot object.

        :type project: :py:class:`~copr.client_v2.resources.Project`
        :param project: parent project for the chroot
        :param str name: chroot name

        :rtype: :py:class:`~copr.client_v2.resources.ProjectChroot`
        """

        url = "{0}/{1}".format(self.get_base_url(project), name)
        response = self.nc.request(url)

        return ProjectChroot.from_response(
            handle=self,
            response=response,
            data_dict=response.json,
            project=project,
        )

    def get_list(self, project):
        """ Retrieves project chroot list object.

        :type project: :py:class:`~copr.client_v2.resources.Project`
        :param project: parent project for the chroot

        :rtype: :py:class:`~copr.client_v2.resources.ProjectChrootList`
        """
        response = self.nc.request(self.get_base_url(project))
        return ProjectChrootList.from_response(
            handle=self,
            response=response,
            project=project
        )

    def disable(self, project, name):
        """ Disables one chroot for the project

        :type project: :py:class:`~copr.client_v2.resources.Project`
        :param project: parent project for the chroot

        :param str name: chroot name to disable
        """
        url = "{0}/{1}".format(self.get_base_url(project), name)
        response = self.nc.request(url, method="DELETE", do_auth=True)
        return OperationResult(self, response)

    def enable(self, project, name, buildroot_pkgs=None):
        """ Enables one chroot for the project

        :type project: :py:class:`~copr.client_v2.resources.Project`
        :param project: parent project for the chroot

        :param str name: chroot name to enable

        :params buildroot_pkgs: packages to add into the buildroot
        :type buildroot_pkgs: list of str

        :rtype: :py:class:`.OperationResult`
        """

        new_entity = ProjectChrootEntity(
            name=name,
            buildroot_pkgs=buildroot_pkgs or list()
        )
        response = self.nc.request(
            self.get_base_url(project),
            method="POST",
            data=new_entity.to_json(),
            do_auth=True
        )
        return OperationResult(self, response)

    def update(self, project, chroot_entity):
        """
        :type project: copr.client_v2.resources.Project
        :param chroot_entity: Entity to update
        :type chroot_entity: :py:class:`.entities.ProjectChrootEntity`

        :rtype: :py:class:`.OperationResult`
        """
        url = "{0}/{1}".format(self.get_base_url(project), chroot_entity.name)
        response = self.nc.request(
            url,
            method="PUT",
            data=chroot_entity.to_json(),
            do_auth=True
        )
        return OperationResult(self, response)


class MockChrootHandle(AbstractHandle):

    def __init__(self, client, nc, root_url, href):
        super(MockChrootHandle, self).__init__(client, nc, root_url)
        self._href = href
        self._base_url = "{0}{1}".format(self.root_url, href)

    def get_base_url(self):
        return self._base_url

    def get_one(self, name):
        """ Retrieves mock chroot object.

        :param str name: chroot name

        :rtype: :py:class:`~copr.client_v2.resources.MockChroot`
        """
        url = "{0}/{1}".format(self.get_base_url(), name)
        response = self.nc.get(url)
        return MockChroot.from_response(
            handle=self,
            response=response,
            data_dict=response.json,
        )

    def get_list(self, active_only=True):
        """ Retrieves mock chroot list object.

        :param bool active_only: when True, shows only chroots which can be used for builds

        :rtype: :py:class:`~copr.client_v2.resources.MockChrootList`
        """
        options = dict(active_only=active_only)

        response = self.nc.get(
            self.get_base_url(),
            query_params=options
        )
        return MockChrootList.from_response(
            handle=self,
            response=response,
            options=options
        )



from .client import CoprClient

from .client_v2.client import CoprClient as ClientV2

create_client2_from_params = ClientV2.create_from_params
create_client2_from_file_config = ClientV2.create_from_file_config



__docformat__ = "restructuredtext en"

from .responses import ProjectWrapper, BuildWrapper, ProjectChrootWrapper


class IParser(object):
    """ Parser interface
    """
    provided_fields = set()

    @staticmethod
    def parse(data, field, **kwargs):
        """
        Creates `value` from field

        :param data: json structure from api
        :param field:

        :return: parsed value
        """
        raise NotImplementedError("Interface doesn't provide fields")


def fabric_simple_fields_parser(fields, name=None):
    class FieldsParser(IParser):
        provided_fields = set(fields)

        @staticmethod
        def parse(data, field, **kwargs):
            if field in FieldsParser.provided_fields:
                if field in data:
                    return data[field]
                else:
                    raise KeyError("Response missing field `{}`".format(field))
            else:
                raise KeyError("Field `{}` not supported by parser".
                               format(field))

    if name:
        FieldsParser.__name__ = str(name)

    return FieldsParser


CommonMsgErrorOutParser = fabric_simple_fields_parser(
    ["output", "message", "error"], "CommonMsgErrorOutParser"
)


class ProjectDetailsFieldsParser(IParser):
    provided_fields = set(["description", "instructions", "last_modified", "name"])

    @staticmethod
    def parse(data, field, **kwargs):
        if field in ProjectDetailsFieldsParser.provided_fields:
            if "detail" in data:
                if field in data["detail"]:
                    return data["detail"][field]
                else:
                    raise KeyError("Response missing field `{}`".format(field))
            else:
                raise KeyError("Response missing `detail` section")
        else:
            raise KeyError("Field `{}` not supported by parser".
                           format(field))


class ProjectChrootsParser(IParser):
    provided_fields = set(["chroots"])

    @staticmethod
    def parse(data, field, client=None, **kwargs):
        if field == "chroots":
            if "detail" in data and "yum_repos" in data["detail"]:
                request_kwargs = kwargs.get("request_kwargs")
                return [
                    ProjectChrootWrapper(
                        client=client,
                        username=request_kwargs["username"],
                        projectname=request_kwargs["projectname"],
                        chrootname=chrootname,
                        repo_url=url
                    )
                    for chrootname, url in data["detail"]["yum_repos"].items()
                ]
            else:
                raise KeyError("Response missing data about chroots")
        else:
            raise KeyError("Field `{}` not supported by parser".
                           format(field))


class ProjectListParser(IParser):
    provided_fields = set(["projects_list"])

    @staticmethod
    def parse(data, field, client=None, **kwargs):
        if field == "projects_list":
            if "repos" in data:
                return [
                    ProjectWrapper(
                        client=client,
                        username=prj.get("username", None),
                        # TODO:, sad api doesn't
                        # include username into the response
                        # of user_projects_list

                        # TODO: inconsistency in api between search ans list
                        # user projects
                        projectname=prj.get("name") or prj.get("coprname"),

                        description=prj.get("description"),
                        yum_repos=prj.get("yum_repos"),
                        additional_repos=prj.get("additional_repos"),
                    )
                    for prj in data["repos"]
                ]
            else:
                raise KeyError("Response missing data about projects")
        else:
            raise KeyError("Field `{}` not supported by parser".
                           format(field))


class NewBuildListParser(IParser):
    provided_fields = set(["builds_list"])

    @staticmethod
    def parse(data, field, client=None, **kwargs):
        if field == "builds_list":
            if "ids" in data:
                request_kwargs = kwargs.get("request_kwargs")
                return [
                    BuildWrapper(
                        client=client,
                        build_id=build_id,
                        username=request_kwargs["username"],
                        projectname=request_kwargs["projectname"],

                    ) for build_id in data["ids"]
                ]
            else:
                raise KeyError("Response missing data about new builds")
        else:
            raise KeyError("Field `{}` not supported by parser".
                           format(field))



# coding: utf-8

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from abc import ABCMeta, abstractproperty

import sys
import os
import logging
import weakref

import six
from six import with_metaclass
from six.moves import configparser

from .resources import Root
from .handlers import ProjectHandle, ProjectChrootHandle, BuildHandle, MockChrootHandle, BuildTaskHandle
from .net_client import NetClient

if sys.version_info < (2, 7):
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
else:
    from logging import NullHandler

from ..exceptions import CoprConfigException, CoprNoConfException
from ..util import UnicodeMixin

log = logging.getLogger(__name__)
log.addHandler(NullHandler())


class HandlersProvider(object, with_metaclass(ABCMeta)):

    @abstractproperty
    def projects(self):
        """
        :rtype: ProjectHandle
        """
        pass

    @abstractproperty
    def project_chroots(self):
        """
        :rtype: ProjectChrootHandle
        """
        pass

    @abstractproperty
    def builds(self):
        """
        :rtype: BuildHandle
        """
        pass

    @abstractproperty
    def build_tasks(self):
        """
        :rtype: BuildTaskHandle
        """
        pass

    @abstractproperty
    def mock_chroots(self):
        """
        :rtype: MockChrootHandle
        """
        pass


class CoprClient(UnicodeMixin, HandlersProvider):
    """ Main interface to the copr service

    :param NetClient net_client: wrapper for http requests
    :param unicode root_url: used as copr projects root
    :param bool no_config: helper flag to indicate that no config was provided

    Could be created:
        - using static method :meth:`~copr.client_v2.client.CoprClient.create_from_file_config`
        - using static method :meth:`~copr.client_v2.client.CoprClient.create_from_params`

    If you create Client directly call :meth:`CoprClient.post_init` method after the creation.
    """

    def __init__(self, net_client, root_url=None, no_config=False):
        """

        """
        self.nc = net_client
        self.root_url = root_url or u"http://copr.fedoraproject.org"

        self.no_config = no_config
        self._post_init_done = False

        self.root = None

        self._projects = None
        self._project_chroots = None
        self._builds = None
        self._build_tasks = None
        self._mock_chroots = None

    def _check_client_init(self):
        if not self._post_init_done:
            raise RuntimeError("CoprClient wasn't initialized, use class-methods "
                               "create_from* to get instance of CoprClient")

    @property
    def projects(self):
        """
        :rtype: :py:class:`~copr.client_v2.handlers.ProjectHandle`
        """
        self._check_client_init()
        return self._projects

    @property
    def project_chroots(self):
        """
        :rtype: :py:class:`~copr.client_v2.handlers.ProjectChrootHandle`
        """
        self._check_client_init()
        return self._project_chroots

    @property
    def builds(self):
        """
        :rtype: :py:class:`~copr.client_v2.handlers.BuildHandle`
        """
        self._check_client_init()
        return self._builds

    @property
    def build_tasks(self):
        """
        :rtype: :py:class:`~copr.client_v2.handlers.BuildTaskHandle`
        """
        self._check_client_init()
        return self._build_tasks

    @property
    def mock_chroots(self):
        """
        :rtype: :py:class:`~copr.client_v2.handlers.MockChrootHandle`
        """
        self._check_client_init()
        return self._mock_chroots

    def __unicode__(self):
        return (
            u"<Copr client. api root url: {}, config provided: {}, net client: {}>"
            .format(self.root_url, not self.no_config, self.nc)
        )

    @property
    def api_root(self):
        """
            Url to API endpoint
        """
        return "{0}/api_2".format(self.root_url)

    @classmethod
    def create_from_params(cls, root_url=None, login=None, token=None):
        """ Create client instance using the given parameters

        :param str root_url: Url to the Copr service, default: "http://copr.fedoraproject.org"
        :param str login: api login
        :param str token: api token

        :rtype: :py:class:`.client_v2.client.CoprClient`
        """
        nc = NetClient(login, token)
        client = cls(nc, root_url, no_config=True)
        client.post_init()
        return client

    @classmethod
    def create_from_file_config(cls, filepath=None, ignore_error=False):
        """
        Creates Copr client using the information from the config file.

        :param filepath: specifies config location,
            default: "~/.config/copr"
        :type filepath: `str`
        :param bool ignore_error: When true and config is missing,
            creates default Client without credentionals

        :rtype: :py:class:`.client_v2.client.CoprClient`
        """
        raw_config = configparser.ConfigParser()
        if not filepath:
            filepath = os.path.join(os.path.expanduser("~"), ".config", "copr")
        config = {}
        if not raw_config.read(filepath):
            log.warning(
                "No configuration file '~/.config/copr' found. "
                "See man copr-cli for more information")
            config["no_config"] = True
            if not ignore_error:
                raise CoprNoConfException()
            else:
                return cls.create_from_params()
        else:
            try:
                for field in ["login", "token", "copr_url"]:
                    if six.PY3:
                        config[field] = raw_config["copr-cli"].get(field, None)
                    else:
                        config[field] = raw_config.get("copr-cli", field, None)
                nc = NetClient(config["login"], config["token"])
                client = cls(nc, root_url=config["copr_url"], )

            except configparser.Error as err:
                if not ignore_error:
                    raise CoprConfigException(
                        "Bad configuration file: {0}".format(err))
                else:
                    return cls.create_from_params()

        client.post_init()
        return client

    def post_init(self):
        """ Finalizes client initialization by querying API root info
        """

        log.debug("Getting root resources")

        response = self.nc.request(self.api_root)

        # obtain api info info
        self.root = Root.from_response(response, self.root_url)

        # instantiating handlers
        self._projects = ProjectHandle(
            weakref.proxy(self), self.nc, root_url=self.root_url,
            projects_href=self.root.get_href_by_name(u"projects"),)
        self._project_chroots = ProjectChrootHandle(
            weakref.proxy(self), self.nc, root_url=self.root_url)

        self._builds = BuildHandle(
            weakref.proxy(self), self.nc, root_url=self.root_url,
            builds_href=self.root.get_href_by_name(u"builds"),)

        self._build_tasks = BuildTaskHandle(
            weakref.proxy(self), self.nc, root_url=self.root_url,
            build_tasks_href=self.root.get_href_by_name("build_tasks")
        )

        self._mock_chroots = MockChrootHandle(
            weakref.proxy(self), self.nc, root_url=self.root_url,
            href=self.root.get_href_by_name(u"mock_chroots")
        )

        self._post_init_done = True



# -*- coding: UTF-8 -*-
# pylint: disable=W1202

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

import json
import sys
import os
import logging

import requests
import six

from six.moves import configparser
from requests_toolbelt.multipart.encoder import (MultipartEncoder,
                                                 MultipartEncoderMonitor)

# urlparse from six is not available on el7
# because it requires at least python-six-1.4.1
if sys.version_info[0] == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

if sys.version_info < (2, 7):
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
else:
    from logging import NullHandler

log = logging.getLogger(__name__)
log.addHandler(NullHandler())

from ..exceptions import CoprConfigException, CoprNoConfException, \
    CoprRequestException, \
    CoprUnknownResponseException

from .responses import ProjectHandle, \
    CoprResponse, BuildHandle, BaseHandle, ProjectChrootHandle

from .parsers import fabric_simple_fields_parser, ProjectListParser, \
    CommonMsgErrorOutParser, NewBuildListParser, ProjectChrootsParser, \
    ProjectDetailsFieldsParser

from ..util import UnicodeMixin

# TODO: add deco to check that login/token are provided
# and  raise correct error
# """ "No configuration file '~/.config/copr' found. "
# "see documentation at /usr/share/doc/python-copr/ "
# """
# or
# """
# "No api login and/or api token are provided"
#    "See man copr-cli for more information")
# """


class CoprClient(UnicodeMixin):
    """ Main interface to the copr service

    :ivar unicode username: username used by default for all requests
    :ivar unicode login: user login, used for identification
    :ivar unicode token: copr api token
    :ivar unicode copr_url: used as copr projects root

    Could be created:
        - directly
        - using static method :py:meth:`CoprClient.create_from_file_config`

    """

    def __init__(self, username=None, login=None, token=None, copr_url=None,
                 no_config=False):
        """
            :param unicode username: username used by default for all requests
            :param unicode login: user login, used for identification
            :param unicode token: copr api token
            :param unicode copr_url: used as copr projects root
            :param bool no_config: helper flag to indicate that no config was provided
        """

        self.token = token
        self.login = login
        self.username = username
        self.copr_url = copr_url or "http://copr.fedoraproject.org/"

        self.no_config = no_config

    def __unicode__(self):
        return (
            u"<Copr client. username: {0}, api url: {1}, login presents: {2}, token presents: {3}>"
            .format(self.username, self.api_url, bool(self.login), bool(self.token))
        )

    @property
    def api_url(self):
        """
            Url to API endpoint
        """
        return "{0}/api".format(self.copr_url)

    @staticmethod
    def create_from_file_config(filepath=None, ignore_error=False):
        """
        Creates Copr client using the information from the config file.

        :param filepath: specifies config location,
            default: "~/.config/copr"
        :type filepath: `str`
        :param bool ignore_error: When true creates default Client
            without credentionals

        :rtype: :py:class:`~.client.CoprClient`

        """

        raw_config = configparser.ConfigParser()
        if not filepath:
            filepath = os.path.join(os.path.expanduser("~"), ".config", "copr")
        config = {}
        if not raw_config.read(filepath):
            log.warning(
                "No configuration file '~/.config/copr' found. "
                "See man copr-cli for more information")
            config["no_config"] = True
            if not ignore_error:
                raise CoprNoConfException()
        else:
            try:
                for field in ["username", "login", "token", "copr_url"]:
                    if six.PY3:
                        config[field] = raw_config["copr-cli"].get(field, None)
                    else:
                        config[field] = raw_config.get("copr-cli", field, None)

            except configparser.Error as err:
                if not ignore_error:
                    raise CoprConfigException(
                        "Bad configuration file: {0}".format(err))
        return CoprClient(**config)

    def _fetch(self, url, data=None, username=None, method=None,
               skip_auth=False, on_error_response=None, headers=None):
        """ Fetches data from server,
        checks response and raises a CoprRequestException with nice error message
        or CoprUnknownResponseException in case of some some error. \n
        Otherwise return unpacked json object.

            :param url: formed url to fetch
            :param data: [optional] serialised data to send
            :param skip_auth: [optional] don't send auth credentials
            :param username: [optional] use alternative username
            :param on_error_response: [optional] function to handle responses
                with bad status code
            :param headers: [optional] custom request headers

            :return: deserialized response
            :rtype: dict
        """

        if method is None:
            method = "get"

        if not username:
            username = self.username

        log.debug("Fetching url: {0}, for login: {1}".format(url, self.login))
        kwargs = {}
        if not skip_auth:
            kwargs["auth"] = (self.login, self.token)
        if data is not None:
            kwargs["data"] = data
        if headers is not None:
            kwargs["headers"] = headers

        if method not in ["get", "post", "head", "delete", "put"]:
            raise Exception("Method {0} not allowed".format(method))

        try:
            response = requests.request(
                method=method.upper(),
                url=url,
                **kwargs
            )
            log.debug("raw response: {0}".format(response.text))
        except requests.ConnectionError as e:
            raise CoprRequestException(e)

        if "<title>Sign in Copr</title>" in response.text:
            raise CoprRequestException("Invalid API token\n")

        if response.status_code > 299 and on_error_response is not None:
            return on_error_response(response)

        if response.status_code == 404:
            log.error("Bad request, URL not found: {0}".
                      format(url))
        elif 400 <= response.status_code < 500:
            log.error("Bad request, raw response body: {0}".
                      format(response.text))
        elif response.status_code >= 500:
            log.error("Server error, raw response body: {0}".
                      format(response.text))

        try:
            output = json.loads(response.text)
        except ValueError:
            raise CoprUnknownResponseException(
                "Unknown response from the server. Code: {0}, raw response:"
                " \n {1}".format(response.status_code, response.text))
        if response.status_code != 200:
            raise CoprRequestException(output["error"])

        if output is None:
            raise CoprUnknownResponseException("No response from the server.")
        return output

    def get_build_details(self, build_id, projectname=None, username=None):
        """ Returns build details.

            :param build_id: Build identifier
            :type build_id: int

            :param projectname: [optional] Copr project name
            :param username: [optional] Copr project owner

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **handle:** :py:class:`~.responses.BuildHandle`
                - text fields: "project", "owner", "status", "results",
                "submitted_on", "started_on", "ended_on",
                "built_pkgs", "src_pkg", "src_version"


        """

        url = "{0}/coprs/build/{1}/".format(
            self.api_url, build_id)

        data = self._fetch(url, skip_auth=True)
        response = CoprResponse(
            client=self,
            method="get_build_details",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                fabric_simple_fields_parser(
                    [
                        "project", "owner", "status", "results",
                        "submitted_on", "started_on", "ended_on",
                        "built_pkgs", "src_pkg", "src_version",
                    ],  # TODO: convert unix time
                    "BuildDetailsParser"
                )
            ]
        )
        response.handle = BuildHandle(
            self, response=response, build_id=build_id,
            projectname=getattr(response, "project", projectname),
            username=getattr(response, "owner", username)
        )
        return response

    def cancel_build(self, build_id, projectname=None, username=None):
        """ Cancels build.
            Auth required.
            If build can't be canceled do nothing.

            :param build_id: Build identifier
            :type build_id: int

            :param projectname: [optional] Copr project name
            :param username: [optional] Copr project owner

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **handle:** :py:class:`~.responses.BuildHandle`
                - text fields: "status"
        """

        url = "{0}/coprs/cancel_build/{1}/".format(
            self.api_url, build_id)

        data = self._fetch(url, skip_auth=False, method='post')
        response = CoprResponse(
            client=self,
            method="cancel_build",
            data=data,
            parsers=[
                fabric_simple_fields_parser(["status", "output", "error"]),
            ]
        )
        response.handle = BuildHandle(
            self, response=response, build_id=build_id,
            projectname=projectname, username=username
        )
        return response

    def create_new_build(self, projectname, pkgs, username=None,
                         timeout=None, memory=None, chroots=None,
                         progress_callback=None):
        """ Creates new build

            :param projectname: name of Copr project (without user namespace)
            :param pkgs: list of packages to include in build
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        if not username:
            username = self.username
        data = {
            "memory_reqs": memory,
            "timeout": timeout
        }

        if urlparse(pkgs[0]).scheme != "":
            api_endpoint = "new_build"
            data["pkgs"] = " ".join(pkgs)
        else:
            try:
                api_endpoint = "new_build_upload"
                f = open(pkgs[0], "rb")
                data["pkgs"] = (os.path.basename(f.name), f, "application/x-rpm")
            except IOError as e:
                raise CoprRequestException(e)

        url = "{0}/coprs/{1}/{2}/{3}/".format(
            self.api_url, username, projectname, api_endpoint
        )

        for chroot in chroots or []:
            data[chroot] = "y"

        m = MultipartEncoder(data)

        callback = progress_callback or (lambda x: x)

        monit = MultipartEncoderMonitor(m, callback)
        data = self._fetch(url, monit, method="post",
                           headers={'Content-Type': monit.content_type})

        response = CoprResponse(
            client=self,
            method="cancel_build",
            data=data,
            request_kwargs={
                "projectname": projectname,
                "username": username
            },
            parsers=[
                CommonMsgErrorOutParser,
                NewBuildListParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname, username=username)

        return response

    def get_project_details(self, projectname, username=None):
        """ Returns project details

            :param projectname: Copr projectname
            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - text fields: "description", "instructions", "last_modified",
                "name"

                - **chroots**: list of
                  :py:class:`~.responses.ProjectChrootWrapper`
        """
        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/detail/".format(
            self.api_url, username, projectname
        )

        data = self._fetch(url, skip_auth=True)
        # return ProjectDetailsResponse(self, response, projectname, username)

        response = CoprResponse(
            client=self,
            method="get_project_details",
            data=data,
            request_kwargs={
                "projectname": projectname,
                "username": username
            },
            parsers=[
                ProjectChrootsParser,
                ProjectDetailsFieldsParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname,
                                        username=username)
        return response

    def delete_project(self, projectname, username=None):
        """ Deletes the entire project.
            Auth required.

            :param projectname: Copr projectname
            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - text fields: "message"
        """

        if not username:
            username = self.username
        url = "{0}/coprs/{1}/{2}/delete/".format(
            self.api_url, username, projectname
        )

        data = self._fetch(
            url, data={"verify": "yes"}, method="post")

        response = CoprResponse(
            client=self,
            method="delete_project",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname,
                                        username=username)
        return response

    def create_project(
            self, projectname, chroots,
            description=None, instructions=None,
            repos=None, initial_pkgs=None
    ):
        """ Creates a new copr project
            Auth required.

            :param projectname: Copr project name
            :param chroots: List of target chroots
            :param description: [optional] Project description
            :param instructions: [optional] Instructions for end users

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectHandle`
                - text fields: "message"
        """

        url = "{0}/coprs/{1}/new/".format(
            self.api_url, self.username)

        if not chroots:
            raise Exception("You should provide chroots")

        if isinstance(chroots, list):
            chroots = [chroots]

        if isinstance(repos, list):
            repos = " ".join(repos)

        if isinstance(initial_pkgs, list):
            initial_pkgs = " ".join(initial_pkgs)

        request_data = {
            "name": projectname,
            "repos": repos,
            "initial_pkgs": initial_pkgs,
            "description": description,
            "instructions": instructions
        }
        for chroot in chroots:
            request_data[chroot] = "y"

        # TODO: def on bad_response()
        result_data = self._fetch(url, data=request_data, method="post")

        response = CoprResponse(
            client=self,
            method="create_project",
            data=result_data,
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname)
        return response

    def modify_project(self, projectname, username=None,
                       description=None, instructions=None,
                       repos=None, disable_createrepo=None):
        """ Modifies main project configuration.
            Auth required.

            :param projectname: Copr project name
            :param username: [optional] use alternative username
            :param description: [optional] project description
            :param instructions: [optional] instructions for end users
            :param repos: [optional] list of additional repos to be used during
                the build process

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectHandle`
                - text fields: "buildroot_pkgs"
        """

        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/modify/".format(
            self.api_url, username, projectname
        )
        data = {}
        if description:
            data["description"] = description
        if instructions:
            data["instructions"] = instructions
        if repos:
            data["repos"] = repos
        if disable_createrepo:
            data["disable_createrepo"] = disable_createrepo

        result_data = self._fetch(url, data=data, method="post")

        response = CoprResponse(
            client=self,
            method="modify_project",
            data=result_data,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectDetailsFieldsParser,
                fabric_simple_fields_parser(["buildroot_pkgs"])
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname)
        return response

    def get_projects_list(self, username=None):
        """ Returns list of projects created by the user

            :param username: [optional] use alternative username


            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **projects_list**: list of
                  :py:class:`~.responses.ProjectWrapper`
        """
        if not username:
            username = self.username

        url = "{0}/coprs/{1}/".format(
            self.api_url, username)
        data = self._fetch(url)
        response = CoprResponse(
            client=self,
            method="get_projects_list",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectListParser,
            ]
        )
        response.handle = BaseHandle(client=self, username=username,
                                     response=response)
        return response

    def get_project_chroot_details(self, projectname,
                                   chrootname, username=None):
        """ Returns details of chroot used in project

            :param projectname: Copr project name
            :param chrootname: chroot name

            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectChrootHandle`
                - text fields: "buildroot_pkgs"
        """
        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/detail/{3}/".format(
            self.api_url, username, projectname, chrootname
        )
        data = self._fetch(url, skip_auth=True)
        response = CoprResponse(
            client=self,
            method="get_project_chroot_details",
            data=data,
            parsers=[
                fabric_simple_fields_parser(
                    ["buildroot_pkgs", "output", "error"],
                    "BuildDetailsParser"
                )
            ]
        )
        response.handle = ProjectChrootHandle(
            client=self, chrootname=chrootname, username=username,
            projectname=projectname, response=response
        )
        return response

    def modify_project_chroot_details(self, projectname, chrootname,
                                      pkgs=None, username=None):
        """ Modifies chroot used in project

            :param projectname: Copr project name
            :param chrootname: chroot name

            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectChrootHandle`
                - text fields: "buildroot_pkgs"

        """
        if pkgs is None:
            pkgs = []

        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/modify/{3}/".format(
            self.api_url, username, projectname, chrootname
        )
        data = {
            "buildroot_pkgs": " ".join(pkgs)
        }
        data = self._fetch(url, data=data, method="post")

        response = CoprResponse(
            client=self,
            method="modify_project_chroot_details",
            data=data,
            parsers=[
                fabric_simple_fields_parser(
                    ["buildroot_pkgs", "output", "error"],
                    "BuildDetailsParser"
                )
            ]
        )
        response.handle = ProjectChrootHandle(
            client=self, chrootname=chrootname, username=username,
            projectname=projectname, response=response
        )
        return response

    def search_projects(self, query):
        """ Search projects by substring

            :param query: substring to search

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **projects_list**: list of
                  :py:class:`~.responses.ProjectWrapper`
        """
        url = "{0}/coprs/search/{1}/".format(
            self.api_url, query
        )
        data = self._fetch(url, skip_auth=True)

        response = CoprResponse(
            client=self,
            method="search_projects",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectListParser
            ]
        )
        response.handle = BaseHandle(client=self, response=response)
        return response



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from copr.client.responses import ProjectWrapper

def test_project_wrapper_unicode():
    project = ProjectWrapper("<client>", "<username>", "<projectname>",
                             description="ěščřžýáíé",
                             instructions="ěščřžýáíé")
    assert str(project)



# coding: utf-8

from marshmallow import Schema, fields
from marshmallow.validate import OneOf

# todo: maybe define schemas for Link sets in each indvidual/collection
# class LinkSchema(Schema):
#     href = fields.Str()
#
from .common import BuiltPackage, ALLOWED_BUILD_STATES


class EmptySchema(Schema):
    pass


class ProjectSchema(Schema):

    id = fields.Int()
    name = fields.Str()

    owner = fields.Str()
    # is_a_group_project = fields.Bool()
    # group = fields.Str(allow_none=True)

    description = fields.Str()
    instructions = fields.Str()
    homepage = fields.Url(allow_none=True)
    contact = fields.Str(allow_none=True)

    disable_createrepo = fields.Bool(allow_none=True)
    build_enable_net = fields.Bool(allow_none=True)
    last_modified = fields.DateTime(dump_only=True)

    repos = fields.List(fields.Str())


class ProjectCreateSchema(ProjectSchema):

    id = fields.Int(allow_none=True)
    chroots = fields.List(fields.Str())


class ProjectChrootSchema(Schema):

    name = fields.Str()
    buildroot_pkgs = fields.List(fields.Str())

    comps = fields.Str(allow_none=True)
    comps_name = fields.Str(allow_none=True)
    comps_len = fields.Int(load_only=True, allow_none=True)


class BuiltPackageSchema(Schema):

    name = fields.Str()
    version = fields.Str()

    # @post_load
    def make_object(self, data):
        return BuiltPackage(**data)


class BuildSchema(Schema):

    id = fields.Int(load_only=True)
    state = fields.Str()

    submitter = fields.Str(load_only=True)

    built_packages = fields.Nested(BuiltPackageSchema, many=True, load_only=True, allow_none=True)

    package_version = fields.Str(load_only=True, allow_none=True)
    package_name = fields.Str(load_only=True, allow_none=True)

    repos = fields.List(fields.Str(), load_only=True)

    submitted_on = fields.Int(load_only=True)
    started_on = fields.Int(load_only=True, allow_none=True)
    ended_on = fields.Int(load_only=True, allow_none=True)

    enable_net = fields.Bool(load_only=True)

    source_type = fields.Str(load_only=True)
    source_metadata = fields.Raw(load_only=True)


class BuildTaskSchema(Schema):

    state = fields.Str(load_only=True, validate=[
        OneOf(ALLOWED_BUILD_STATES)
    ])
    started_on = fields.Int(load_only=True, allow_none=True)
    ended_on = fields.Int(load_only=True, allow_none=True)
    git_hash = fields.Str(load_only=True, allow_none=True)
    chroot_name = fields.Str(load_only=True)
    build_id = fields.Int(load_only=True)
    result_dir_url = fields.Str(load_only=True, allow_none=True)


class MockChrootSchema(Schema):

    name = fields.Str(load_only=True)
    os_release = fields.Str(load_only=True)
    os_version = fields.Str(load_only=True)
    arch = fields.Str(load_only=True)
    is_active = fields.Bool(load_only=True)
